// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "frameXmlWinEdit.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2003-2004
// Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.
//
// -----------------------------------------------------------------------
//
// Acknowledgement: Parts of the code, as indicated below, are taken from
// the EditAppDemos project, written by Michael Hieke for the SynEdit
// component suite (see "http://SynEdit.SourceForge.net").

unit frameXmlWinEdit;

interface

uses
  dkAdomCore, dkUriUtils, dkWideStringUtils, dkLangUtils, uXmlEditTypes,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Types,
  Dialogs, ComCtrls, ActnList, SynEditMiscClasses, SynEditSearch,
  SynEdit, SynEditHighlighter, SynHighlighterXML, SynEditTypes, SynEditPrint,
  AutoListUtils;

type
  TXmlWinEditFrame = class(TFrame)
    SynEditSearch: TSynEditSearch;
    StatusBar: TStatusBar;
    SynEdit: TSynEdit;
    procedure SynEditReplaceText(Sender: TObject; const ASearch,
      AReplace: string; Line, Column: Integer; var Action: TSynReplaceAction);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

  TXmlWinEdit = class(TXmlWinEditFrame, IXmlEditCtrl)
  private
    FSynEditPrint: TSynEditPrint;
    FSearchReplaceDlgLeft: Integer; // Backup for the Dialog position.
    FSearchReplaceDlgTop: Integer;  // Backup for the Dialog position.

    function CanFocusStdcall: Boolean; stdcall;
    function GetData: TBytes; stdcall;
    function GetEditorFont: TFont;
    function GetModified: Boolean; stdcall;
    procedure SetData(const Value: TBytes);
    procedure SetEditorFont(const Value: TFont);
    procedure SetFocusStdcall; stdcall;
    procedure SetModified(const Value: Boolean);  stdcall;
    procedure SetSynEditPrint(const Value: TSynEditPrint);
  protected
    function GetCaretXY: TBufferCoord; virtual;
    function GetSearchTextFromControl: WideString; virtual;
    procedure Notification(AComponent: TComponent; operation: TOperation); override;
    procedure SetCaretXY(const Value: TBufferCoord); virtual;
    function ShowSearchReplaceDialog(var Params: TXmlSearchParams;
                                         AReplace: Boolean): Boolean; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    function CanCopy: Boolean; virtual; stdcall;
    function CanCut: Boolean; virtual; stdcall;
    function CanDelete: Boolean; virtual; stdcall;
    function IXmlEditCtrl.CanFocus = CanFocusStdcall;
    function CanFocus: Boolean; override;
    function CanPaste: Boolean; virtual; stdcall;
    function CanRedo: Boolean; virtual; stdcall;
    function CanReplace: Boolean; virtual; stdcall;
    function CanUndo: Boolean; virtual; stdcall;
    function CaretPosStr: string; virtual; stdcall;
    procedure CopyToClipboard; virtual; stdcall;
    procedure CutToClipboard; virtual; stdcall;
    procedure DeleteSelection; virtual; stdcall;
    function GetEditorState: string; virtual; stdcall;
    function IsEmpty: Boolean; virtual; stdcall;
    procedure PasteFromClipboard; virtual; stdcall;
    procedure Print; virtual; stdcall;
    procedure PrintPreview(const PrinterSetupAction: TAction); virtual; stdcall;
    procedure Redo; virtual; stdcall;
    procedure SearchReplace(const ASearch, AReplace: WideString;
      AOptions: TSynSearchOptions); virtual; stdcall;
    procedure SelectAll; virtual; stdcall;
    function SetCaret(const ByteOffset, Line, Column: Integer): Boolean; virtual; stdcall;
    procedure IXmlEditCtrl.SetFocus = SetFocusStdcall;
    procedure SetFocus; override;
    function StartReplace(var Params: TXmlSearchParams): Boolean; virtual; stdcall;
    function StartSearch(var Params: TXmlSearchParams): Boolean; virtual; stdcall;
    procedure Undo; virtual; stdcall;

    property CaretXY: TBufferCoord read GetCaretXY write SetCaretXY;
    property Modified: Boolean read GetModified write SetModified;
    property Data: TBytes read GetData write SetData;
  published
    property EditorFont: TFont read GetEditorFont write SetEditorFont;
    property SynEditPrint: TSynEditPrint read FSynEditPrint write SetSynEditPrint;
  end;

procedure Register;

implementation

uses
  Math, SynEditMiscProcs, dlgPrintPreview, dlgSearchText,
  dlgReplaceText, dlgConfirmReplace;

resourcestring
  SInsert = 'Insert';
  SOverwrite = 'Overwrite';
  SReadOnly = 'Read Only';

  SSearchStringNotFound = 'Search string ''%s'' not found.';

procedure Register;
begin
  RegisterComponents('XML Editor', [ TXmlWinEdit ]);
end;

{$R *.dfm}

{ TXmlWinEditFrame }

procedure TXmlWinEditFrame.SynEditReplaceText(Sender: TObject; const ASearch,
  AReplace: string; Line, Column: Integer; var Action: TSynReplaceAction);
// Modified version of code taken from the SearchReplaceDemo
// project, written by Michael Hieke for the SynEdit
// component suite (see "http://SynEdit.SourceForge.net").
var
  APos: TPoint;
  ConfirmReplaceDialog: TConfirmReplaceDialog;
  EditRect: TRect;
begin
  if ASearch = AReplace then
    Action := raSkip
  else begin
    APos := SynEdit.ClientToScreen(
      SynEdit.RowColumnToPixels(
      SynEdit.BufferToDisplayPos(
        BufferCoord(Column, Line) ) ) );
    EditRect := ClientRect;
    EditRect.TopLeft := ClientToScreen(EditRect.TopLeft);
    EditRect.BottomRight := ClientToScreen(EditRect.BottomRight);

    ConfirmReplaceDialog := TConfirmReplaceDialog.Create(Application);
    try
      ConfirmReplaceDialog.PrepareShow(EditRect, APos.X, APos.Y,
        APos.Y + SynEdit.LineHeight, ASearch);
      case ConfirmReplaceDialog.ShowModal of
        mrYes: Action := raReplace;
        mrYesToAll: Action := raReplaceAll;
        mrNo: Action := raSkip;
        else Action := raCancel;
      end;
    finally
      ConfirmReplaceDialog.Free;
    end;

  end;
end;

{ TXmlWinEdit }

function TXmlWinEdit.CanCopy: Boolean;
begin
  Result := SynEdit.SelAvail;
end;

function TXmlWinEdit.CanCut: Boolean;
begin
  Result := SynEdit.SelAvail and not SynEdit.ReadOnly;
end;

function TXmlWinEdit.CanDelete: Boolean;
begin
  Result := SynEdit.SelAvail and not SynEdit.ReadOnly;
end;

function TXmlWinEdit.CanFocus: Boolean;
var
  Form: TCustomForm;
begin
  Result := SynEdit.CanFocus;

  // The following code is a workaround for a bug in the Delphi VCL:
  if Result and not (csDesigning in ComponentState) then
  begin
    Form := GetParentForm(Self);
    if Assigned(Form)
      then Result := Form.Enabled and Form.Visible
      else Result := False;
  end;
end;

function TXmlWinEdit.CanFocusStdcall: Boolean;
begin
  Result := CanFocus;
end;

function TXmlWinEdit.CanPaste: Boolean;
begin
  Result := SynEdit.CanPaste;
end;

function TXmlWinEdit.CanRedo: Boolean;
begin
  Result := SynEdit.CanRedo;
end;

function TXmlWinEdit.CanReplace: Boolean;
begin
  Result := not SynEdit.ReadOnly;
end;

function TXmlWinEdit.CanUndo: Boolean;
begin
  Result := SynEdit.CanUndo;
end;

function TXmlWinEdit.CaretPosStr: string;
begin
  with CaretXY do
    if (Char > 0) and (Line > 0)
      then Result := Format(' %6d:%3d ', [Line, Char])
      else Result := '';
end;

procedure TXmlWinEdit.CopyToClipboard;
begin
  SynEdit.CopyToClipboard;
end;

constructor TXmlWinEdit.Create(AOwner: TComponent);
begin
  inherited;
  FSearchReplaceDlgLeft := -1;
  FSearchReplaceDlgTop := -1;
  StatusBar.Panels[0].Text := 'Screen encoding: UTF-16';
  StatusBar.Panels[1].Text := 'Output encoding: UTF-8';
end;

procedure TXmlWinEdit.CutToClipboard;
begin
  SynEdit.CutToClipboard;
end;

procedure TXmlWinEdit.DeleteSelection;
begin
  SynEdit.SelText := '';
end;

function TXmlWinEdit.GetCaretXY: TBufferCoord;
begin
  Result := SynEdit.CaretXY;
end;

function TXmlWinEdit.GetData: TBytes;
var
  L: Integer;
begin
  Result := WideBytesOf(SynEdit.Text);
end;

function TXmlWinEdit.GetEditorFont: TFont;
begin
  Result := SynEdit.Font;
end;

function TXmlWinEdit.GetEditorState: string;
begin
  if SynEdit.ReadOnly then
    Result := SReadOnly
  else if SynEdit.InsertMode then
    Result := SInsert
  else
    Result := SOverwrite;
end;

function TXmlWinEdit.GetModified: Boolean;
begin
  Result := SynEdit.Modified;
end;

function TXmlWinEdit.GetSearchTextFromControl: WideString;
begin
  if SynEdit.SelAvail and (SynEdit.BlockBegin.Char = SynEdit.BlockEnd.Char)   // xxx Change Char to Line?
    then Result := SynEdit.SelText
    else Result := SynEdit.GetWordAtRowCol(SynEdit.CaretXY);
end;

function TXmlWinEdit.IsEmpty: Boolean;
var
  I: Integer;
begin
  Result := True;
  with SynEdit do
    for I := 0 to Pred(Lines.Count) do
      if Lines[I] <> '' then begin
        Result := False;
        Exit;
      end;
end;

procedure TXmlWinEdit.Notification(AComponent: TComponent; operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FSynEditPrint) then
    FSynEditPrint := nil;
end;

procedure TXmlWinEdit.PasteFromClipboard;
begin
  SynEdit.PasteFromClipboard;
end;

procedure TXmlWinEdit.Print;
begin
  FSynEditPrint.Print;
end;

procedure TXmlWinEdit.PrintPreview(const PrinterSetupAction: TAction);
var
  PrintPreviewDlg: TPrintPreviewDlg;
begin
  PrintPreviewDlg := TPrintPreviewDlg.Create(Application);
  try
    with PrintPreviewDlg do begin
      SynEditPrintPreview.SynEditPrint := FSynEditPrint;
      ShowModal;
    end;
  finally
    PrintPreviewDlg.Free;
  end;
end;

procedure TXmlWinEdit.Redo;
begin
  with SynEdit do begin
    BeginUpdate;
    try
      Redo;
    finally
      EndUpdate;
    end;
  end;
end;

procedure TXmlWinEdit.SearchReplace(const ASearch, AReplace: WideString;
  AOptions: TSynSearchOptions);
begin
  if SynEdit.SearchReplace(ASearch, AReplace, AOptions) = 0 then begin
    MessageBeep(MB_ICONASTERISK);
    ShowMessage(Format(SSearchStringNotFound, [ASearch]));
    if ssoBackwards in AOptions then
      SynEdit.BlockEnd := SynEdit.BlockBegin
    else
      SynEdit.BlockBegin := SynEdit.BlockEnd;
    SynEdit.CaretXY := SynEdit.BlockBegin;
  end;
end;

procedure TXmlWinEdit.SelectAll;
begin
  SynEdit.SelectAll;
end;

function TXmlWinEdit.SetCaret(const ByteOffset, Line, Column: Integer): Boolean;
begin
  if Line > 0 then begin
    CaretXY := TBufferCoord(Point(Max(Column, 1), Line)); // Remark: TBufferCoord and TPoint are compatible.
    Result := True;
  end else
    Result := False;
end;

procedure TXmlWinEdit.SetCaretXY(const Value: TBufferCoord);
begin
  SynEdit.CaretXY := Value;
end;

procedure TXmlWinEdit.SetFocus;
begin
  SynEdit.SetFocus;
end;

procedure TXmlWinEdit.SetFocusStdcall;
begin
  SetFocus;
end;

procedure TXmlWinEdit.SetModified(const Value: Boolean);
begin
  SynEdit.Modified := Value;
end;

procedure TXmlWinEdit.SetSynEditPrint(const Value: TSynEditPrint);
begin
  if FSynEditPrint = Value then Exit;
  if Assigned(FSynEditPrint)
    then FSynEditPrint.RemoveFreeNotification(Self);
  FSynEditPrint := Value;
  if Assigned(Value) then begin
    Value.SynEdit := SynEdit;
    Value.FreeNotification(Self);
  end;
end;

procedure TXmlWinEdit.SetData(const Value: TBytes);
begin
  SynEdit.Text := WideStringOf(Value);
end;

procedure TXmlWinEdit.SetEditorFont(const Value: TFont);
begin
  SynEdit.Font := Value;
end;

function TXmlWinEdit.ShowSearchReplaceDialog(var Params: TXmlSearchParams;
  AReplace: Boolean): Boolean;
// Modified version of code taken from the EditAppDemos
// project, written by Michael Hieke for the SynEdit
// component suite (see "http://SynEdit.SourceForge.net").
var
  Dlg: TTextSearchDialog;
begin
  if AReplace
    then Dlg := TTextReplaceDialog.Create(Self)
    else Dlg := TTextSearchDialog.Create(Self);
  with Dlg do try
    // Assign backup screen position of the Search/Replace Dialog
    if (FSearchReplaceDlgLeft > -1) and (FSearchReplaceDlgTop > -1) then begin
      Left := FSearchReplaceDlgLeft;
      Top := FSearchReplaceDlgTop;
      Position := poDesigned;
    end;
    // Assign search options
    SearchBackwards := Params.SearchBackwards;
    SearchCaseSensitive := Params.SearchCaseSensitive;
    SearchFromCursor := Params.SearchFromCursor;
    SearchInSelectionOnly := SynEdit.SelAvail;
    // Start with last search text or selected text or text at caret, respectively
    SearchText := Params.SearchText;
    if Params.SearchTextAtCaret then
      SearchText := GetSearchTextFromControl;
      { TODO 4 -cRefactor : Return value of GetSearchTextFromControl is WideString! }
    SearchTextHistory := Params.SearchTextHistory;
    if AReplace then with Dlg as TTextReplaceDialog do begin
      ReplaceText := Params.ReplaceText;
      ReplaceTextHistory := Params.ReplaceTextHistory;
    end;
    SearchWholeWords := Params.SearchWholeWords;
    if ShowModal = mrOK then begin
      Params.SearchBackwards := SearchBackwards;
      Params.SearchCaseSensitive := SearchCaseSensitive;
      Params.SearchFromCursor := SearchFromCursor;
      Params.SearchSelectionOnly := SearchInSelectionOnly;
      Params.SearchWholeWords := SearchWholeWords;
      Params.SearchText := SearchText;
      Params.SearchTextHistory := SearchTextHistory;
      if AReplace then with dlg as TTextReplaceDialog do begin
        Params.ReplaceText := ReplaceText;
        Params.ReplaceTextHistory := ReplaceTextHistory;
      end;
      Params.SearchFromCaret := Params.SearchFromCursor;
      Result := True;
    end else Result := False;
    FSearchReplaceDlgLeft := Dlg.Left;
    FSearchReplaceDlgTop := Dlg.Top;
  finally
    Dlg.Free;
  end;
end;

function TXmlWinEdit.StartReplace(var Params: TXmlSearchParams): Boolean;
begin
  Result := ShowSearchReplaceDialog(Params, True);
end;

function TXmlWinEdit.StartSearch(var Params: TXmlSearchParams): Boolean;
begin
  Result := ShowSearchReplaceDialog(Params, False);
end;

procedure TXmlWinEdit.Undo;
begin
  with SynEdit do begin
    BeginUpdate;
    try
      Undo;
    finally
      EndUpdate;
    end;
  end;
end;

end.
