// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "uVersionCheck.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2004
// Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

unit uVersionCheck;

interface

uses
  Classes;

type
  TVersionCheckerResult = ( vrNewVersion, vrNoNewVersion, vrNetworkError);

  TVersionCheckerCompleted = procedure(Sender: TObject;
                                       VCResult: TVersionCheckerResult) of object;

  TVersionCheckerMessageOpts = set of ( voNewVersionMsg, voNoNewVersionMsg, voErrorMsg );

  TVersionChecker = class(TComponent)
  private
    FOptions: TVersionCheckerMessageOpts;
    FUserAgent: string;
    FOnCompleted: TVersionCheckerCompleted;
    procedure ThreadTerminateHandler(Sender: TObject);
  public
    procedure CheckVersion(const FirstAttempt: Boolean);

  published
    property Options: TVersionCheckerMessageOpts read FOptions write FOptions;
    property UserAgent: string read FUserAgent write FUserAgent;

    property OnCompleted: TVersionCheckerCompleted read FOnCompleted write FOnCompleted;
  end;

{$IFDEF CLR}

  TVersionDetail = (
    vdAll,
    vdVersion,
    vdMajorVersion,
    vdMinorVersion,
    vdBuild,
    vdPrivate
  );

function GetFileVersionStr(const FileName: string;
  const VersionDetail: TVersionDetail = vdAll): string; platform;

{$ELSE}

  TVersionDetail = (
    vdAll,
    vdVersion,
    vdMajorVersion,
    vdMinorVersion,
    vdRelease,
    vdBuild
  );

function GetFileVersionStr(const FileName: string;
  const VersionDetail: TVersionDetail = vdAll): string; platform;

{$ENDIF}

procedure Register;

implementation

uses
{$IFDEF CLR}
  System.Diagnostics,
{$ELSE}
  Windows,
{$ENDIF}
  dlgPreferences, AdomNetwork_5_1, SysUtils, Dialogs, Forms;

resourcestring
  SNewVersionURL = 'http://www.philo.de/xmledit/version.txt';

  SNetworkError = 'A network error occurred.';
  SNewVersionAvailable =  'A new version of the Open XML Editor is available for download at ''http://www.philo.de/xmledit/''.';
  SNoNewVersionAvailable =  'There is no new version of the Open XML Editor available.';

procedure Register;
begin
  RegisterComponents('XML Editor', [ TVersionChecker ]);
end;

{ Helper functions }

{$IFDEF CLR}
function GetFileVersionStr(const FileName: string;
  const VersionDetail: TVersionDetail = vdAll): string;
var
  VersionInfo: FileVersionInfo;
begin
  VersionInfo := FileVersionInfo.GetVersionInfo(FileName);
  case VersionDetail of
    vdAll:
      begin
        if Assigned(VersionInfo.FileVersion) then
          Result := VersionInfo.FileVersion
        else
          Result := '0.0.0.0';
      end;
    vdVersion:
      Result := Format('%d.%d', [ VersionInfo.FileMajorPart, VersionInfo.FileMinorPart ]);
    vdMajorVersion:
      Result := Format('%d', [ VersionInfo.FileMajorPart ]);
    vdMinorVersion:
      Result := Format('%d', [ VersionInfo.FileMinorPart ]);
    vdBuild:
      Result := Format('%d', [ VersionInfo.FileBuildPart ]);
    vdPrivate:
      Result := Format('%d', [ VersionInfo.FilePrivatePart ]);
  end;
end;

{$ELSE}

function GetFileVersionStr(const FileName: string;
  const VersionDetail: TVersionDetail = vdAll): string;
var
  Dummy: DWORD;
  InfoBufferSize: DWORD;
  FixedFileInfo: PVSFixedFileInfo;
  VersionBuffer: Pointer;
  VersionSize: DWORD;
begin
  Result := '';
  InfoBufferSize := GetFileVersionInfoSize(PChar(FileName), Dummy);
  if InfoBufferSize <> 0 then
  begin
    GetMem(VersionBuffer, InfoBufferSize);
    try
      if GetFileVersionInfo(PChar(FileName), Dummy, InfoBufferSize, VersionBuffer) then
        if VerQueryValue(VersionBuffer, '\', Pointer(FixedFileInfo), VersionSize) then
        begin
          case VersionDetail of
            vdAll:
              Result := Format('%d.%d.%d.%d',
                [ FixedFileInfo.dwFileVersionMS shr 16 and $0000FFFF,
                  FixedFileInfo.dwFileVersionMS and $0000FFFF,
                  FixedFileInfo.dwFileVersionLS shr 16 and $0000FFFF,
                  FixedFileInfo.dwFileVersionLS and $0000FFFF ]);
            vdVersion:
              Result := Format('%d.%d',
                [ FixedFileInfo.dwFileVersionMS shr 16 and $0000FFFF,
                  FixedFileInfo.dwFileVersionMS and $0000FFFF ]);
            vdMajorVersion:
              Result := Format('%d',
                [ FixedFileInfo.dwFileVersionMS shr 16 and $0000FFFF ]);
            vdMinorVersion:
              Result := Format('%d',
                [ FixedFileInfo.dwFileVersionMS and $0000FFFF ]);
            vdRelease:
              Result := Format('%d',
                [ FixedFileInfo.dwFileVersionLS shr 16 and $0000FFFF ]);
            vdBuild:
              Result := Format('%d',
                [ FixedFileInfo.dwFileVersionLS and $0000FFFF ]);
          end;
        end;
    finally
      FreeMem(VersionBuffer);
    end;
  end;
end;
{$ENDIF}

procedure TVersionChecker.CheckVersion(const FirstAttempt: Boolean);
var
  CheckVersionThread: THttpThread;
  VersionStr: string;
begin
  VersionStr := GetFileVersionStr(Application.ExeName, vdAll);

  // Retrieve version info:
  CheckVersionThread := THttpThread.Create(True);
  try
    with CheckVersionThread do begin

      FreeOnTerminate := True;
      URL := SNewVersionURL;
      with ProxyParams do
        with UserPreferences do
          if UseProxies then begin
            ProxyServer := HttpProxy;
            ProxyPassword := HttpProxyPassword;
            ProxyPort := HttpProxyPort;
            ProxyUsername :=  HttpProxyUsername;
          end else begin
            ProxyServer := '';
            ProxyPassword := '';
            ProxyPort := 0;
            ProxyUsername :=  '';
          end;
      if FirstAttempt
        then Request.UserAgent := Trim(FUserAgent) + '/' + VersionStr + '-new'
        else Request.UserAgent := Trim(FUserAgent) + '/' + VersionStr;
      OnTerminate := ThreadTerminateHandler;

      Resume;

    end;
  except
    CheckVersionThread.Free;
  end;
end;

procedure TVersionChecker.ThreadTerminateHandler(Sender: TObject);
begin
  if Sender is THttpThread then
    with THttpThread(Sender) do
      if Status = hsCompleted then begin
        if Trim(Data) <> GetFileVersionStr(Application.ExeName, vdAll) then begin
          if voNewVersionMsg in Options then
            ShowMessage(SNewVersionAvailable)
        end else begin
          if voNoNewVersionMsg in Options then
            ShowMessage(SNoNewVersionAvailable);
        end;
      end else if voErrorMsg in Options then
        ShowMessage(SNetworkError);
end;

end.
