// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "uXmlDocMan.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2004-2005
// Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

unit uXmlDocMan;

interface

uses
  dkAdomCore, AdomErrorReporting_5_1, AdomTranscoder_5_1,
  SysUtils, Classes;

type
  TXmlDocManagerStates = set of (
    msHexTextInvalid,
    msWinTextInvalid
  );

  TXmlDynamicError = class(TDomError)
  private
    function RNodeUserDataAsInteger(const Key: WideString): Integer;
  protected
    function GetEndByteNumber: Int64; override;
    function GetEndCharNumber: Int64; override;
    function GetEndColumnNumber: Int64; override;
    function GetEndLineNumber: Int64; override;
    function GetStartByteNumber: Int64; override;
    function GetStartCharNumber: Int64; override;
    function GetStartColumnNumber: Int64; override;
    function GetStartLineNumber: Int64; override;
    function GetUri: WideString; override;
  public
    function CloneError: TDomError; override;
  end;

  TXmlDynamicErrorAutoStrings = class(TDomErrorAutoStrings)
  protected
    function Get(Index: Integer): string; override;
    function GetError(Index: Integer): TXmlDynamicError; reintroduce; virtual;
  public
    function AddError(AError: TXmlDynamicError): Integer; reintroduce; overload; virtual;
    function InsertError(Index: Integer; AError: TXmlDynamicError): Boolean; reintroduce; overload; virtual;
    procedure SendStateChangedNotification; virtual;

    property Errors[Index: Integer]: TXmlDynamicError read GetError;
  end;

  TXmlDocManager = class(TComponent)
  private
    FDocumentTree: TDomDocument;
    FDocumentUri: WideString;
    FDomImplementation: TDomImplementation;
    FDomToXmlParser: TDomToXmlParser;
    FEnglishErrorTranslator: TDomEnglishErrorTranslator;
    FErrorAutoStrings: TXmlDynamicErrorAutoStrings;
    FIsWellformed: Boolean;
    FMimeType: string;
    FRawData: TBytes;
    FRawDataEncoding: WideString;
    FXmlToDomParser: TXmlToDomParser;

    function GetDocumentTree: TDomDocument;
    function GetResourceResolver: TCustomResourceResolver;
    procedure SetDocumentUri(const Value: WideString);
    procedure SetResourceResolver(const Value: TCustomResourceResolver);

    function Transcode(const Source: TBytes;
                       const SourceEncoding,
                             TargetEncoding: WideString;
                       const XmlDeclOption: TXmlTranscodeOpt): TBytes;

    procedure ErrorEventHandler(Sender: TObject; Error: TdomError;
      var Go: Boolean);
    procedure ParserAfterWriteEventHandler(Sender: TXmlStreamBuilder;
      PieceType: TdomPieceType; const Locator: IDomLocator);
    procedure ParserBeforeWriteEventHandler(Sender: TXmlStreamBuilder;
      PieceType: TdomPieceType; const Locator: IDomLocator);

  protected
    property DocumentTree: TDomDocument read GetDocumentTree;
  public
    constructor Create(AOwner: TComponent); override;
    function CheckValidity: Boolean; virtual;
    function GetUnicodeData: TBytes;  // Transcodes RawData to UTF-16LE without BOM or XML-declaration.
    function HasErrorStrings: Boolean; virtual;
    function SerializeData(const Encoding: WideString;
                           const XmlDeclOption: TXmlTranscodeOpt;
                           const UseByteOrderMark: TXmlBOMOpt): TBytes;
    procedure SetRawData(const ARawData: TBytes;
                         const AMimeType: string;
                         const ARawDataEncoding: WideString;
                         const IncDecl: Boolean);

    property ErrorAutoStrings: TXmlDynamicErrorAutoStrings read FErrorAutoStrings;
    property IsWellformed: Boolean read FIsWellformed;
    property MimeType: string read FMimeType;
    property RawData: TBytes read FRawData;
    property RawDataEncoding: WideString read FRawDataEncoding;
  published
    property DocumentUri: WideString read FDocumentUri write SetDocumentUri;
    property ResourceResolver: TCustomResourceResolver read GetResourceResolver write SetResourceResolver;
  end;

{ Helper methods }

procedure GetNodeStartPosition(Node: TdomNode; out Line, Column: Int64);

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('XML Editor', [ TXmlDocManager ]);
end;

{ Helper methods }
procedure GetNodeStartPosition(Node: TdomNode; out Line, Column: Int64);
var
  PColumn: Pointer;
  PLine: Pointer;
begin
  if Assigned(Node) then begin
    PLine := Node.GetUserData('StartLine');
    PColumn := Node.GetUserData('StartColumn');
    if Assigned(PLine)
      then Line := Integer(PLine)
      else Line := -1;
    if Assigned(PColumn)
      then Column := Integer(PColumn)
      else Column := -1;
  end else begin
    Line := -1;
    Column := -1;
  end;
end;

{ TXmlDynamicError }

function TXmlDynamicError.CloneError: TDomError;
begin
  Result := TXmlDynamicError.Create(RelatedException, StartByteNumber,
          StartCharNumber, StartColumnNumber, StartLineNumber, StartTabsInLine,
          EndByteNumber, EndCharNumber, EndColumnNumber, EndLineNumber,
          EndTabsInLine, Uri, RelatedDtdObject, RelatedNode, Code, Clue);
end;

function TXmlDynamicError.GetEndByteNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('EndByteNumber');
    if Result = -1 then
      Result := inherited GetEndCharNumber;
  end else Result := inherited GetEndCharNumber;
end;

function TXmlDynamicError.GetEndCharNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('EndCharNumber');
    if Result = -1 then
      Result := inherited GetEndCharNumber;
  end else Result := inherited GetEndCharNumber;
end;

function TXmlDynamicError.GetEndColumnNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('EndColumn');
    if Result = -1 then
      Result := inherited GetEndColumnNumber;
  end else Result := inherited GetEndColumnNumber;
end;

function TXmlDynamicError.GetEndLineNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('EndLine');
    if Result = -1 then
      Result := inherited GetEndLineNumber;
  end else Result := inherited GetEndLineNumber;
end;

function TXmlDynamicError.GetStartByteNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('StartByte');
    if Result = -1 then
      Result := inherited GetStartByteNumber;
  end else Result := inherited GetStartCharNumber;
end;

function TXmlDynamicError.GetStartCharNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('StartChar');
    if Result = -1 then
      Result := inherited GetStartCharNumber;
  end else Result := inherited GetStartCharNumber;
end;

function TXmlDynamicError.GetStartColumnNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('StartColumn');
    if Result = -1 then
      Result := inherited GetStartColumnNumber;
  end else Result := inherited GetStartColumnNumber;
end;

function TXmlDynamicError.GetStartLineNumber: Int64;
begin
  if Assigned(RelatedNode) then begin
    Result := RNodeUserDataAsInteger('StartLine');
    if Result = -1 then
      Result := inherited GetStartLineNumber;
  end else Result := inherited GetStartLineNumber;
end;

function TXmlDynamicError.GetUri: WideString;
begin
  if Assigned(RelatedNode) then begin
    if Assigned(RelatedNode.ownerDocument)
      then Result := RelatedNode.OwnerDocument.DocumentUri
      else Result := inherited GetURI;
  end else Result := inherited GetURI;
end;

function TXmlDynamicError.RNodeUserDataAsInteger(
  const Key: WideString): Integer;
var
  P: Pointer;
begin
  if Assigned(RelatedNode) then begin
    P := RelatedNode.GetUserData(Key);
    if Assigned(P)
      then Result := Integer(P)
      else Result := -1;
  end else
    Result := -1;
end;

{ TXmlDynamicErrorAutoStrings }

function TXmlDynamicErrorAutoStrings.AddError(
  AError: TXmlDynamicError): Integer;
begin
  Result := inherited AddError(AError);
end;

function TXmlDynamicErrorAutoStrings.Get(Index: Integer): string;
begin
  Result := inherited Get(Index);

  with Errors[Index] do
    if Assigned(RelatedNode) then begin
      Result := Concat(Result, ': [', RelatedNode.NodeName,']');
    end else
      if Assigned(RelatedDtdObject) then
        Result := Concat(Result, ' -- DTD object name: ', RelatedDtdObject.Name);
end;

function TXmlDynamicErrorAutoStrings.GetError(
  Index: Integer): TXmlDynamicError;
begin
  Result := Objects[Index] as TXmlDynamicError;
end;

function TXmlDynamicErrorAutoStrings.InsertError(Index: Integer;
  AError: TXmlDynamicError): Boolean;
begin
  Result := inherited InsertError(Index, AError);
end;

procedure TXmlDynamicErrorAutoStrings.SendStateChangedNotification;
begin
  DoNotifyStateChanged;
end;

{ TXmlDocManager }

constructor TXmlDocManager.Create(AOwner: TComponent);
begin
  inherited;
  FDocumentUri := '';

  FMimeType := '';

  FRawData := nil;
  FRawDataEncoding := '';

  FDocumentTree := nil;
  FIsWellformed := True;

  FDomImplementation := TDomImplementation.Create(Self);
  FDomToXmlParser := TDomToXmlParser.Create(Self);
  FXmlToDomParser := TXmlToDomParser.Create(Self);
  FErrorAutoStrings := TXmlDynamicErrorAutoStrings.Create(Self);
  FEnglishErrorTranslator := TDomEnglishErrorTranslator.Create(Self);

  FDomImplementation.OnError := ErrorEventHandler;

  FEnglishErrorTranslator.TabWidth := 1;

  FXmlToDomParser.DOMImpl := FDomImplementation;

  with FDomToXmlParser do begin
    DOMImpl := FDomImplementation;
    ErrorReportingEnabled := False;  // Switched off, so no serialization errors are reported in message panel.
    UseActiveCodePage := False;

    OnAfterWrite := ParserAfterWriteEventHandler;
    OnBeforeWrite := ParserBeforeWriteEventHandler;
  end;

  FErrorAutoStrings.ErrorTranslator := FEnglishErrorTranslator;
end;

function TXmlDocManager.CheckValidity: Boolean;
begin
  if IsWellformed then begin
    FErrorAutoStrings.DisableControls;
    try
      FErrorAutoStrings.Clear;
      if Assigned(FDocumentTree)
        then Result := FDocumentTree.ValidationAgent.ValidateDocument(erExpand)
        else Result := True;
    finally
      FErrorAutoStrings.EnableControls;
    end;
  end else
    Result := False;
end;

procedure TXmlDocManager.ErrorEventHandler(Sender: TObject; Error: TdomError;
  var Go: Boolean);
var
  NewError: TXmlDynamicError;
begin
  NewError := TXmlDynamicError.CreateFromLocator(Error.RelatedException, Error, Error.Code, Error.Clue);
  try
    FErrorAutoStrings.AddError(NewError);
  finally
    NewError.Free;
  end;
end;

function TXmlDocManager.GetDocumentTree: TDomDocument;
begin
  Result := FDocumentTree;
end;

function TXmlDocManager.GetResourceResolver: TCustomResourceResolver;
begin
  Result := FDomImplementation.ResourceResolver;
end;

function TXmlDocManager.GetUnicodeData: TBytes;
begin
  Result := Transcode(RawData, RawDataEncoding, 'UTF-16LE', toNoXmlDecl);
end;

function TXmlDocManager.HasErrorStrings: Boolean;
begin
  Result := FErrorAutoStrings.Count > 0;
end;

procedure TXmlDocManager.ParserAfterWriteEventHandler(Sender: TXmlStreamBuilder;
  PieceType: TdomPieceType; const Locator: IDomLocator);
begin
  with Locator do
    if Assigned(RelatedNode) then
      if PieceType in [ xmlProcessingInstruction, xmlComment, xmlCDATA,
                        xmlPCDATA, xmlDoctype, xmlEndTag, xmlEmptyElementTag,
                        xmlEntityRef, xmlAttribute, xmlXmlDeclaration,
                        xmlTextDeclaration ] then begin
        RelatedNode.SetUserData('EndByte', Pointer(Sender.ByteCount), nil);
        RelatedNode.SetUserData('EndColumn', Pointer(Sender.ColumnCount + 1), nil);
        RelatedNode.SetUserData('EndLine', Pointer(Sender.LineFeedCount + 1), nil);
      end;
end;

procedure TXmlDocManager.ParserBeforeWriteEventHandler(Sender: TXmlStreamBuilder;
  PieceType: TdomPieceType; const Locator: IDomLocator);
begin
  with Locator do
    if Assigned(RelatedNode) then
      if PieceType in [ xmlProcessingInstruction, xmlComment, xmlCDATA,
                        xmlPCDATA, xmlDoctype, xmlStartTag, xmlEmptyElementTag,
                        xmlEntityRef, xmlAttribute, xmlXmlDeclaration,
                        xmlTextDeclaration ] then begin
        RelatedNode.SetUserData('StartByte', Pointer(Sender.ByteCount), nil);
        RelatedNode.SetUserData('StartColumn', Pointer(Sender.ColumnCount + 1), nil);
        RelatedNode.SetUserData('StartLine', Pointer(Sender.LineFeedCount + 1), nil);
      end;
end;

function TXmlDocManager.SerializeData(const Encoding: WideString;
                                      const XmlDeclOption: TXmlTranscodeOpt;
                                      const UseByteOrderMark: TXmlBOMOpt): TBytes;
var
  OutputStream: TMemoryStream;
begin
  if RawData = nil then begin
    Result := nil;
    Exit;
  end;

  if MimeType = 'application/octet-stream' then begin
    Result := RawData;
    Exit;
  end;

  if DocumentTree <> nil then begin

    try
      FDomToXmlParser.IncludeXmlDecl := XmlDeclOption in [toEnforceXmlDecl, toCopyXmlDecl];
      OutputStream := TMemoryStream.Create();
      try
        FDomToXmlParser.UseByteOrderMark := UseByteOrderMark;
        if FDomToXmlParser.WriteToStream(DocumentTree, Encoding, OutputStream) then begin
          OutputStream.Position := 0;
          SetLength(Result, OutputStream.Size);
          OutputStream.ReadBuffer(Result[0], OutputStream.Size);
        end else

          { TODO 3 -cRefactory : Replace with exception?! }

          // The document tree contains an undisplayable and unescapable character.
          // So we must prepare the Hex display:
          { TODO 5 -cRefactory : Check Remark! }
          Result := Transcode(RawData, RawDataEncoding, Encoding, XmlDeclOption);
      finally
        OutputStream.Free;
      end;
    except
      Result := Transcode(RawData, RawDataEncoding, Encoding, XmlDeclOption);
    end;

  end else
    Result := Transcode(RawData, RawDataEncoding, Encoding, XmlDeclOption);

end;

procedure TXmlDocManager.SetDocumentUri(const Value: WideString);
begin
  FDocumentUri := Value;
end;

procedure TXmlDocManager.SetRawData(const ARawData: TBytes;
  const AMimeType: string; const ARawDataEncoding: WideString;
  const IncDecl: Boolean);
var
  ByteStream: TBytesStream;
  InputSource: TXmlInputSource;
begin
  if CompareText(AMimeType, 'application/xml') = 0 then begin

    if (FRawData <> ARawData) or
       (FMimeType <> AMimeType) or
       (FRawDataEncoding <> ARawDataEncoding)  { TODO 5 -cRevisit : Check also for equivalent names? (Like 'latin1' vs. 'ISO-8859-1'.) }
    then begin
      FMimeType := 'application/xml';
      FRawData := ARawData;
      FErrorAutoStrings.Clear;
      FreeAndNil(FDocumentTree);
      FIsWellformed := True;
      if FRawData = nil then Exit;

      // Parse document and determine wellformedness:
      FErrorAutoStrings.DisableControls;
      try
        FErrorAutoStrings.Clear;
        try
          ByteStream := TBytesStream.Create(FRawData);
          try
            InputSource := TXmlInputSource.Create(ByteStream, '', DocumentUri, 1024, ARawDataEncoding, False, 0, 0, 0, 0, 1);
            try
              FDocumentTree := FXmlToDomParser.Parse(InputSource);
              FRawDataEncoding := FDocumentTree.InputEncoding;
            finally
              InputSource.Free;
            end;
          finally
            ByteStream.Free;
          end;
        except
          FreeAndNil(FDocumentTree);
          FIsWellformed := False;
          FRawData := ARawData;
          FRawDataEncoding := ARawDataEncoding;
        end;
      finally
        FErrorAutoStrings.EnableControls;
      end;
    end;

  end else if CompareText(AMimeType, 'application/octet-stream') = 0 then begin

    FMimeType := 'application/octet-stream';
    FRawData := ARawData;
    FRawDataEncoding := '';
    FErrorAutoStrings.Clear;
    FreeAndNil(FDocumentTree);
    FIsWellformed := False;

  end else
    raise ENot_Supported_Err.CreateFmt('MIME type "%s" not supported.', [AMimeType]);

end;

procedure TXmlDocManager.SetResourceResolver(
  const Value: TCustomResourceResolver);
begin
  FDomImplementation.ResourceResolver := Value;
end;

function TXmlDocManager.Transcode(const Source: TBytes;
  const SourceEncoding, TargetEncoding: WideString;
  const XmlDeclOption: TXmlTranscodeOpt): TBytes;
var
  InputStream: TBytesStream;
  OutputStream: TMemoryStream;
  XmlTranscoder: TXmlTranscoder;
begin
  XmlTranscoder := nil;
  OutputStream := nil;
  InputStream := TBytesStream.Create(Source);
  try
    OutputStream := TMemoryStream.Create();
    XmlTranscoder := TXmlTranscoder.Create(InputStream, OutputStream, 4096, SourceEncoding, False);
    with XmlTranscoder do begin
      OutputEncoding := TargetEncoding;
      XmlDeclOpt := XmlDeclOption;
      Transcode;
      OutputStream.Position := 0;
      SetLength(Result, OutputStream.Size);
      OutputStream.ReadBuffer(Result[0], OutputStream.Size);
    end;
  finally
    XmlTranscoder.Free;
    OutputStream.Free;
    InputStream.Free;
  end;
end;

end.
