// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "uXmlTabCtrls.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2004
// Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.
//
// -----------------------------------------------------------------------
//
// Acknowledgement: Parts of the code, as indicated below, are taken from
// the EditAppDemos project, written by Michael Hieke for the SynEdit
// component suite (see "http://SynEdit.SourceForge.net").

unit uXmlTabCtrls;

interface

uses
  dkAdomCore, AdomNetwork_5_1, frameXmlWinEdit, frameXmlEdit,
  uXmlEditTypes, dlgPageSetup, AutoListUtils, IdHTTP,
  AdomErrorReporting_5_1, SynEditTypes, SynEditPrint, SynHighlighterXML,
  Graphics, Messages, ActnList, Buttons, Classes, ComCtrls, Controls, Dialogs,
  SysUtils;

type
  TXmlEditSaved = procedure(Sender: TObject; const OldUri, NewUri: WideString)
    of object;

  TXmlEditPageControl = class;

  TXmlEditTabSheet = class(TTabSheet)
  private
    FTimeStamp: TDateTime;
    FUntitledNumber: Integer;
    FXmlEditFrame: TXmlEditFrame;
    function GetCaption: string;
    procedure SetCaption(const Value: string);
    procedure UpdateUntitledNumber;
  protected
    function GetDocumentUri: WideString; virtual;
    function GetEditorFont: TFont; virtual;
    function GetEditorState: string; virtual;
    function GetErrorStrings: TDomErrorAutoStrings; virtual;
    function GetExternalHexEditor: TFilename; virtual;
    function GetFileName: TFileName; virtual;
    function GetIsWellformed: Boolean; virtual;
    function GetMimeType: string; virtual;
    function GetModified: Boolean; virtual;
    function GetPageControl: TXmlEditPageControl; virtual;
    procedure SetDocumentUri(const AUri: WideString); virtual;
    procedure SetEditorFont(const Value: TFont); virtual;
    procedure SetExternalHexEditor(const Value: TFilename); virtual;
    procedure SetPageControl(const APageControl: TXmlEditPageControl); virtual;
    procedure SetTimeStamp(const Value: TDateTime); virtual;
    procedure SetupHexEditorRequestEventHandler(Sender: TObject);
  public
    constructor Create(      AOwner: TComponent;
                       const ASourceCode: TBytes;
                       const ADocumentUri: WideString;
                       const AModified: Boolean;
                       const ATimeStamp: TDateTime;
                       const AEditorFont: TFont;
                       const AExternalHexEditor: TFilename); reintroduce;
    destructor Destroy; override;

    procedure Activate; virtual;
    function AskSaveChanges: Integer; virtual;
    function CanCheckValidity: Boolean; virtual;
    function CanCheckWellformedness: Boolean; virtual;
    function CanCopy: Boolean; virtual;
    function CanCut: Boolean; virtual;
    function CanDelete: Boolean; virtual;
    function CanPaste: Boolean; virtual;
    function CanPrint: Boolean; virtual;
    function CanRedo: Boolean; virtual;
    function CanReplace: Boolean; virtual;
    function CanSave: Boolean; virtual;
    function CanSaveAs: Boolean; virtual;
    function CanSearch: Boolean; virtual;
    function CanSelectAll: Boolean; virtual;
    function CanUndo: Boolean; virtual;
    function CaretPosStr: string; virtual;
    function CheckValidity: Boolean; virtual;
    function CheckWellformedness: Boolean; virtual;
    procedure CopyToClipboard; virtual;
    procedure CutToClipboard; virtual;
    procedure DeleteSelection; virtual;
    procedure FocusEditor; virtual;
    function HasErrorStrings: Boolean; virtual;
    function IsEmpty: Boolean; virtual;
    procedure LoadFromBytes(const Data: TBytes;
                            const AModified: Boolean); virtual;
    procedure PasteFromClipboard; virtual;
    procedure Print(const ASynEditPrint: TSynEditPrint); virtual;
    procedure PrintPreview(const ASynEditPrint: TSynEditPrint;
                           const PrinterSetupAction: TAction); virtual;
    procedure Redo; virtual;
    function Save: Boolean; virtual;
    procedure SearchReplace(const ASearch, AReplace: WideString;
      AOptions: TSynSearchOptions); virtual;
    procedure SelectAll; virtual;
    function SetCaret(const Location: IDomLocator): Boolean; virtual;
    function StartReplace(var Params: TXmlSearchParams): Boolean; virtual;
    function StartSearch(var Params: TXmlSearchParams): Boolean; virtual;
    procedure Undo; virtual;

    property ErrorStrings: TDomErrorAutoStrings read GetErrorStrings;
    property Caption: string read GetCaption write SetCaption;
    property DocumentUri: WideString read GetDocumentUri write SetDocumentUri;
    property EditorFont: TFont read GetEditorFont write SetEditorFont;
    property EditorState: string read GetEditorState;
    property ExternalHexEditor: TFilename read GetExternalHexEditor write SetExternalHexEditor;
    property FileName: TFileName read GetFileName;
    property IsWellformed: Boolean read GetIsWellformed;
    property MimeType: string read GetMimeType;
    property Modified: Boolean read GetModified;
    property PageControl: TXmlEditPageControl read GetPageControl write SetPageControl;
    property TimeStamp: TDateTime read FTimeStamp write SetTimeStamp;
  end;

  TXmlEditPageControl = class(TPageControl)
  private
    FEditorFont: TFont;
    FExternalHexEditor: TFilename;
    FSaxon: TFilename;
    FSaxonProcessCount: Cardinal;

    FCloseBtn: TSpeedButton;

    FResourceResolver: TCustomResourceResolver;
    FErrorStringIterator: TUtilsStringIterator;

    FSaveDialog: TSaveDialog;

    FPageSetupDlg: TPageSetupDlg;
    FPrintDialog: TPrintDialog;
    FPrinterSetupAction: TAction;
    FSynEditPrint: TSynEditPrint;
    FUntitledNumbers: TBits;
    FXmlHighlighter: TSynXmlSyn;

    FSearchParams: TXmlSearchParams;

    FInSetActivePage: Boolean;

    FStartEmpty: Boolean;

    FOnEditorChange: TNotifyEvent;
    FOnSaved: TXmlEditSaved;
    FOnSaxonLaunching: TNotifyEvent;
    FOnSetupHexEditorRequest: TNotifyEvent;
    FOnTabChange: TNotifyEvent;    // Occurs when the active tab has changed.
                                   // (This differs from OnChange which occurs
                                   // only if the user selects another tab.)
                                   // WARNING: This event is sometimes fired
                                   // twice, because the privacy police of the
                                   // base class TPageControl is too restrictive.

    function GetActivePage: TXmlEditTabSheet;
    function GetActivePageIndex: Integer;
    function GetEditorFont: TFont;
    function GetExternalHexEditor: TFilename;
    function GetPage(Index: Integer): TXmlEditTabSheet;
    function GetSaxon: TFilename;
    procedure SetActivePage(const Value: TXmlEditTabSheet);
    procedure SetActivePageIndex(const Value: Integer);
    function IndexOfPage(const SysId: WideString): Integer;
    procedure SetEditorFont(const Value: TFont);
    procedure SetExternalHexEditor(const Value: TFilename);
    procedure SetSaxon(const Value: TFilename);
  protected
    procedure CloseBtnClickHandler(Sender: TObject); virtual;
    procedure DoSaved(const OldUri, NewUri: WideString); virtual;
    procedure DoSaxonLaunching; virtual;
    procedure DoSearchReplaceText(AReplace: Boolean; ABackwards: Boolean); virtual;
    procedure DoSetupHexEditorRequest; virtual;
    procedure DoTabChange; virtual;
    function GetUntitledNumber: Integer; virtual;
    procedure Notification(AComponent: TComponent; operation: TOperation); override;
    procedure ReleaseUntitledNumber(ANumber: Integer); virtual;
    function RequestSaveFileName(var AFilename: TFilename): Boolean; virtual;
    function ResolveResource(var PublicId,
                                 SystemId: WideString): TBytes; virtual;
    function SavePage(const Page: TXmlEditTabSheet): Boolean; virtual;
    function SavePageAs(const Page: TXmlEditTabSheet): Boolean; virtual;
    procedure SetStartEmpty(const Value: Boolean); virtual;
    procedure ShowControl(AControl: TControl); override;
    procedure UpdateActivePage; override;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // Status checks
    function CanCheckValidity: Boolean; virtual;
    function CanCheckWellformedness: Boolean; virtual;
    function CanCloseAll: Boolean; virtual;
    function CanCopy: Boolean; virtual;
    function CanCut: Boolean; virtual;
    function CanDelete: Boolean; virtual;
    function CanPaste: Boolean; virtual;
    function CanPrint: Boolean; virtual;
    function CanRedo: Boolean; virtual;
    function CanReplace: Boolean; virtual;
    function CanSave: Boolean; virtual;
    function CanSaveAs: Boolean; virtual;
    function CanSearch: Boolean; virtual;
    function CanSelectAll: Boolean; virtual;
    function CanUndo: Boolean; virtual;
    function HasActivePage: Boolean; virtual;
    function SaxonIsBusy: Boolean; virtual;

    // XML conformity checks
    function CheckValidity: Boolean; virtual;
    function CheckWellformedness: Boolean; virtual;

    // Tools operations
    procedure LaunchSaxon; virtual;

    // Printing
    procedure PageSetup;
    procedure Print;
    procedure PrintPreview;

    // Edit operations
    procedure CopyToClipboard; virtual;
    procedure CutToClipboard; virtual;
    procedure DeleteSelection; virtual;
    procedure PasteFromClipboard; virtual;
    procedure Redo; virtual;
    procedure SelectAll; virtual;
    procedure Undo; virtual;

    // Search/Replace operations
    procedure Find; virtual;
    procedure FindNext; virtual;
    procedure FindPrev; virtual;
    procedure Replace; virtual;

    // Focus and caret operations
    function ActivatePage(const SysId: WideString): Boolean; virtual;
    function CaretPosStr: string; virtual;
    procedure FocusEditor; virtual;
    function SetCaret(const Location: IDomLocator): Boolean; virtual;

    // I/O operations
    procedure CheckRefreshActivePage;
    function OpenUri(PubId,
                     SysId: WideString): Boolean; virtual;
    function Reload: Boolean; virtual;
    procedure Save; virtual;
    procedure SaveAs; virtual;

    // Page addition, removal and content replacement
    procedure AddEmptyDoc; virtual;
    procedure AddDoc(const Data: TBytes;
                     const SysId: WideString;
                     const Modified: Boolean); virtual;
    procedure Close; virtual;
    procedure CloseAll; virtual;

    property ActivePageIndex: Integer read GetActivePageIndex
      write SetActivePageIndex;
    property ErrorStringIterator: TUtilsStringIterator read FErrorStringIterator;
    property Pages[Index: Integer]: TXmlEditTabSheet read GetPage;
    property SearchParams: TXmlSearchParams read FSearchParams write FSearchParams;
  published
    property ActivePage: TXmlEditTabSheet read GetActivePage write SetActivePage;
    property EditorFont: TFont read GetEditorFont write SetEditorFont;
    property ExternalHexEditor: TFilename read GetExternalHexEditor write SetExternalHexEditor;
    property PrinterSetupAction: TAction read FPrinterSetupAction write FPrinterSetupAction;
    property ResourceResolver: TCustomResourceResolver read FResourceResolver write FResourceResolver;
    property Saxon: TFilename read GetSaxon write SetSaxon;
    property StartEmpty: Boolean read FStartEmpty write SetStartEmpty default True;
    property XmlHighlighter: TSynXmlSyn read FXmlHighlighter;                        // xxx Revisit: Make external?

    property OnEditorChange: TNotifyEvent read FOnEditorChange write FOnEditorChange;
    property OnSaved: TXmlEditSaved read FOnSaved write FOnSaved;
    property OnSaxonLaunching: TNotifyEvent read FOnSaxonLaunching write FOnSaxonLaunching;
    property OnSetupHexEditorRequest: TNotifyEvent read FOnSetupHexEditorRequest write FOnSetupHexEditorRequest;
    property OnTabChange: TNotifyEvent read FOnTabChange write FOnTabChange;
  end;

procedure Register;

implementation

{$R uXmlTabCtrls.res}

uses
  Forms, Windows, Math,
  dkUriUtils, dkWideStringUtils, AppLauncher,
  dlgConfirmReplace, dlgXsltOutput, dlgPrintPreview, dlgXsltError;

resourcestring
  SAskReload = 'Another application has updated file'#13#10#13#10 +
               '"%s"'#13#10#13#10 +
               'Reload it?';

  SAskSaveChanges = 'The text in the "%s" file has changed.'#13#10#13#10 +
                    'Do you want to save the modifications?';

  SNameAlreadyInuse = 'Cannot save with the name of a file'#13#10#13#10 +
                      'that is open in another tab sheet';

  SNonameFileTitle = 'Untitled';

  SFOpenError = 'Internal error: Resource Resolver not assigned.';

  SFilterMaskAllFiles = 'All files|*.*';
  SFilterMaskXmlFiles = 'XML files (*.xml; *.xsd; *.xsl; *.xslt)|*.xml;*.xsd;*.xsl;*.xslt';

  SDocumentIsEmpty = 'The input document is empty.';

  SMustBeSaved = 'XSLT processing operates on saved files only.'#13#10#13#10 +
                 'Please first save your modified XML document.';

var
  // Message IDs:
  WM_CLOSE_TAB: OleVariant; // Used for pending closing opperations of Tab Sheets.

procedure Register;
begin
  RegisterComponents('XML Editor', [ TXmlEditPageControl ]);
end;

{ TXmlEditTabSheet }

procedure TXmlEditTabSheet.Activate;
begin
  if Assigned(PageControl) then
    PageControl.ActivePage := Self;
end;

function TXmlEditTabSheet.AskSaveChanges: Integer;
// Return values:
//            0 There is not enough memory to create the message box.
//   IDCancel 2 The user chose the Cancel button.
//   IDYes    6 The user chose the Yes button.
//   IDNo     7 The user chose the No button, or the text was not modified
//              and thus does not need to be saved.
const
  MBType = MB_YESNOCANCEL or MB_ICONQUESTION;
var
  S: string;
begin
  if Modified then begin

    Activate;
    MessageBeep(MB_ICONQUESTION);

    if FUntitledNumber = -1 then begin
      S := Format(SAskSaveChanges, [FileName]);
    end else
      S := Format(SAskSaveChanges, [Caption]);

    Result := Application.MessageBox(PChar(S), PChar(Application.Title), MBType);

  end else

    Result := IDNo;
end;

function TXmlEditTabSheet.CanCheckValidity: Boolean;
begin
  Result := FXmlEditFrame.CanCheckValidity;
end;

function TXmlEditTabSheet.CanCheckWellformedness: Boolean;
begin
  Result := FXmlEditFrame.CanCheckWellformedness;
end;

function TXmlEditTabSheet.CanCopy: Boolean;
begin
  Result := FXmlEditFrame.CanCopy;
end;

function TXmlEditTabSheet.CanCut: Boolean;
begin
  Result := FXmlEditFrame.CanCut;
end;

function TXmlEditTabSheet.CanDelete: Boolean;
begin
  Result := FXmlEditFrame.CanDelete;
end;

function TXmlEditTabSheet.CanPaste: Boolean;
begin
  Result := FXmlEditFrame.CanPaste;
end;

function TXmlEditTabSheet.CanPrint: Boolean;
begin
  Result := FXmlEditFrame.CanPrint;
end;

function TXmlEditTabSheet.CanRedo: Boolean;
begin
  Result := FXmlEditFrame.CanRedo;
end;

function TXmlEditTabSheet.CanReplace: Boolean;
begin
  Result := FXmlEditFrame.CanReplace;
end;

function TXmlEditTabSheet.CanSave: Boolean;
begin
  Result := FXmlEditFrame.CanSave;
end;

function TXmlEditTabSheet.CanSaveAs: Boolean;
begin
  Result := FXmlEditFrame.CanSaveAs;
end;

function TXmlEditTabSheet.CanSearch: Boolean;
begin
  Result := FXmlEditFrame.CanSearch;
end;

function TXmlEditTabSheet.CanSelectAll: Boolean;
begin
  Result := FXmlEditFrame.CanSelectAll;
end;

function TXmlEditTabSheet.CanUndo: Boolean;
begin
  Result := FXmlEditFrame.CanUndo;
end;

function TXmlEditTabSheet.CaretPosStr: string;
begin
  Result := FXmlEditFrame.CaretPosStr;
end;

function TXmlEditTabSheet.CheckValidity: Boolean;
begin
  Result := FXmlEditFrame.CheckValidity;
end;

function TXmlEditTabSheet.CheckWellformedness: Boolean;
begin
  Result := FXmlEditFrame.CheckWellformedness;
end;

procedure TXmlEditTabSheet.CopyToClipboard;
begin
  FXmlEditFrame.CopyToClipboard;
end;

constructor TXmlEditTabSheet.Create(AOwner: TComponent;
  const ASourceCode: TBytes; const ADocumentUri: WideString;
  const AModified: Boolean; const ATimeStamp: TDateTime;
  const AEditorFont: TFont; const AExternalHexEditor: TFilename);
begin
  inherited Create(AOwner);

  FXmlEditFrame := TXmlEditFrame.Create(Self, ASourceCode, ADocumentUri, AModified);
  FXmlEditFrame.Align := alClient;
  FXmlEditFrame.Parent := Self;
  FXmlEditFrame.OnSetupHexEditorRequest := SetupHexEditorRequestEventHandler;

  FUntitledNumber := -1;
  UpdateUntitledNumber;

  FTimeStamp := ATimeStamp;

  SetEditorFont(AEditorFont);
  SetExternalHexEditor(AExternalHexEditor);
end;

procedure TXmlEditTabSheet.SetupHexEditorRequestEventHandler(Sender: TObject);
begin
  if Assigned(PageControl) then
    PageControl.DoSetupHexEditorRequest;
end;

procedure TXmlEditTabSheet.CutToClipboard;
begin
  FXmlEditFrame.CutToClipboard;
end;

procedure TXmlEditTabSheet.DeleteSelection;
begin
  FXmlEditFrame.DeleteSelection;
end;

destructor TXmlEditTabSheet.Destroy;
begin
  PageControl := nil;
  inherited;
end;

procedure TXmlEditTabSheet.FocusEditor;
begin
  FXmlEditFrame.FocusEditor;
end;

function TXmlEditTabSheet.GetCaption: string;

  function ExtractFileNameFromUri(const AUri: WideString): WideString;  // xxx Change to string!
  var
    I: Integer;
  begin
    I := LastDelimiter('/', AUri);
    Result := Copy(AUri, I + 1, MaxInt);
  end;

begin
  if FUntitledNumber = -1
    then Result := ExtractFileNameFromUri(DocumentUri)
    else Result := SNonameFileTitle + IntToStr(FUntitledNumber);
end;

function TXmlEditTabSheet.GetDocumentUri: WideString;
begin
  Result := FXmlEditFrame.DocumentUri;
end;

function TXmlEditTabSheet.GetEditorFont: TFont;
begin
  Result := FXmlEditFrame.EditorFont;
end;

function TXmlEditTabSheet.GetEditorState: string;
begin
  Result := FXmlEditFrame.EditorState;
end;

function TXmlEditTabSheet.GetErrorStrings: TDomErrorAutoStrings;
begin
  Result := FXmlEditFrame.ErrorAutoStrings;
end;

function TXmlEditTabSheet.GetExternalHexEditor: TFilename;
begin
  Result := FXmlEditFrame.ExternalHexEditor;
end;

function TXmlEditTabSheet.GetFileName: TFileName;
var
  Authority, Query, Fragment: WideString; // dummies
begin
  UriWideStrToFilename(DocumentUri, Result, Authority, Query, Fragment);
end;

function TXmlEditTabSheet.GetIsWellformed: Boolean;
begin
  Result := FXmlEditFrame.IsWellformed;
end;

function TXmlEditTabSheet.GetMimeType: string;
begin
  Result := FXmlEditFrame.MimeType;
end;

function TXmlEditTabSheet.GetModified: Boolean;
begin
  Result := FXmlEditFrame.Modified;
end;

function TXmlEditTabSheet.GetPageControl: TXmlEditPageControl;
begin
  Result := inherited PageControl as TXmlEditPageControl;
end;

function TXmlEditTabSheet.HasErrorStrings: Boolean;
begin
  Result := FXmlEditFrame.HasErrorStrings;
end;

function TXmlEditTabSheet.IsEmpty: Boolean;
begin
  Result := FXmlEditFrame.IsEmpty;
end;

procedure TXmlEditTabSheet.LoadFromBytes(const Data: TBytes;
                                         const AModified: Boolean);
begin
  FXmlEditFrame.LoadFromBytes(Data, AModified);
end;

procedure TXmlEditTabSheet.PasteFromClipboard;
begin
  FXmlEditFrame.PasteFromClipboard;
end;

procedure TXmlEditTabSheet.Print(const ASynEditPrint: TSynEditPrint);
begin
  FXmlEditFrame.Print(ASynEditPrint);
end;

procedure TXmlEditTabSheet.PrintPreview(const ASynEditPrint: TSynEditPrint;
  const PrinterSetupAction: TAction);
begin
  FXmlEditFrame.PrintPreview(ASynEditPrint, PrinterSetupAction);
end;

procedure TXmlEditTabSheet.Redo;
begin
  FXmlEditFrame.Redo;
end;

function TXmlEditTabSheet.Save: Boolean;
begin
  Result := FXmlEditFrame.Save;
end;

procedure TXmlEditTabSheet.SearchReplace(const ASearch,
  AReplace: WideString; AOptions: TSynSearchOptions);
begin
  FXmlEditFrame.SearchReplace(ASearch, AReplace, AOptions);
end;

procedure TXmlEditTabSheet.SelectAll;
begin
  FXmlEditFrame.SelectAll;
end;

procedure TXmlEditTabSheet.SetCaption(const Value: string);
begin
  // Do nothing.
end;

function TXmlEditTabSheet.SetCaret(const Location: IDomLocator): Boolean;
begin
  Result := FXmlEditFrame.SetCaret(Location);
end;

procedure TXmlEditTabSheet.SetDocumentUri(const AUri: WideString);
begin
  FXmlEditFrame.DocumentUri := AUri;
  UpdateUntitledNumber;
end;

procedure TXmlEditTabSheet.SetEditorFont(const Value: TFont);
begin
  FXmlEditFrame.EditorFont := Value;
end;

procedure TXmlEditTabSheet.SetExternalHexEditor(const Value: TFilename);
begin
  FXmlEditFrame.ExternalHexEditor := Value;
end;

procedure TXmlEditTabSheet.SetPageControl(
  const APageControl: TXmlEditPageControl);
var
  OldPageControl: TXmlEditPageControl;
begin
  if PageControl <> APageControl then begin
    OldPageControl := PageControl;
    if Assigned(OldPageControl) and (FUntitledNumber <> -1) then
      PageControl.ReleaseUntitledNumber(FUntitledNumber);

    inherited PageControl := APageControl;
    if Assigned(APageControl) then begin
      FXmlEditFrame.XmlHighlighter := APageControl.XmlHighlighter;
      FXmlEditFrame.ResourceResolver := APageControl.ResourceResolver;
    end else begin
      FXmlEditFrame.XmlHighlighter := nil;
      FXmlEditFrame.ResourceResolver := nil;
    end;
    UpdateUntitledNumber;

    if Assigned(OldPageControl) then
      OldPageControl.DoTabChange;
  end;
end;

procedure TXmlEditTabSheet.SetTimeStamp(const Value: TDateTime);
begin
  FTimeStamp := Value;
end;

function TXmlEditTabSheet.StartReplace(var Params: TXmlSearchParams): Boolean;
begin;
  Result := FXmlEditFrame.StartReplace(Params);
end;

function TXmlEditTabSheet.StartSearch(var Params: TXmlSearchParams): Boolean;
begin
  Result := FXmlEditFrame.StartSearch(Params);
end;

procedure TXmlEditTabSheet.Undo;
begin
  FXmlEditFrame.Undo;
end;

procedure TXmlEditTabSheet.UpdateUntitledNumber;
begin
  if not Assigned(PageControl) then
    Exit;
  if DocumentUri = '' then begin
    if FUntitledNumber = -1 then
      FUntitledNumber := PageControl.GetUntitledNumber;
  end else begin
    if FUntitledNumber <> -1 then begin
      PageControl.ReleaseUntitledNumber(FUntitledNumber);
      FUntitledNumber := -1;
    end;
  end;
  inherited Caption := GetCaption;
end;

{ TXmlEditPageControl }

// Printing

procedure TXmlEditPageControl.PageSetup;
begin
  FPageSetupDlg.SetValues(FSynEditPrint);
  if FPageSetupDlg.ShowModal = mrOk then
    FPageSetupDlg.GetValues(FSynEditPrint);
end;

procedure TXmlEditPageControl.Print;
begin
  if Assigned(ActivePage) then
    if FPrintDialog.Execute then
      ActivePage.Print(FSynEditPrint);
end;

procedure TXmlEditPageControl.PrintPreview;
begin
  if Assigned(ActivePage) then
    ActivePage.PrintPreview(FSynEditPrint, FPrinterSetupAction);
end;

// Search/Replace operations

procedure TXmlEditPageControl.Find;
begin
  if Assigned(ActivePage) then
    if ActivePage.StartSearch(FSearchParams) then
      if FSearchParams.SearchText <> '' then begin
        DoSearchReplaceText(False, FSearchParams.SearchBackwards);
        FSearchParams.SearchFromCaret := True;
        FSearchParams.SearchSelectionOnly := False;
      end;
end;

procedure TXmlEditPageControl.FindNext;
begin
  if Assigned(ActivePage) then
    DoSearchReplaceText(False, False);
end;

procedure TXmlEditPageControl.FindPrev;
begin
  if Assigned(ActivePage) then
    DoSearchReplaceText(False, True);
end;

procedure TXmlEditPageControl.Replace;
begin
  if Assigned(ActivePage) then
    if ActivePage.StartReplace(FSearchParams) then
      if FSearchParams.SearchText <> '' then begin
        DoSearchReplaceText(True, FSearchParams.SearchBackwards);
        FSearchParams.SearchFromCaret := True;
        FSearchParams.SearchSelectionOnly := False;
      end;
end;

// Edit operations

procedure TXmlEditPageControl.CopyToClipboard;
begin
  if Assigned(ActivePage) then
    ActivePage.CopyToClipboard;
end;

procedure TXmlEditPageControl.CutToClipboard;
begin
  if Assigned(ActivePage) then
    ActivePage.CutToClipboard;
end;

procedure TXmlEditPageControl.PasteFromClipboard;
begin
  if Assigned(ActivePage) then
    ActivePage.PasteFromClipboard;
end;

procedure TXmlEditPageControl.DeleteSelection;
begin
  if Assigned(ActivePage) then
    ActivePage.DeleteSelection;
end;

procedure TXmlEditPageControl.SelectAll;
begin
  if Assigned(ActivePage) then
    ActivePage.SelectAll;
end;

procedure TXmlEditPageControl.Redo;
begin
  if Assigned(ActivePage) then
    ActivePage.Redo;
end;

procedure TXmlEditPageControl.Undo;
begin
  if Assigned(ActivePage) then
    ActivePage.Undo;
end;

// Standard methods

function TXmlEditPageControl.IndexOfPage(const SysId: WideString): Integer;
// Returns the index of the page with the specified system identifier, or
// -1 if none was found.
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to Pred(PageCount) do
    if CompareWideStr(Pages[I].DocumentUri, SysId) = 0 then begin
      Result := I;
      Exit;
    end;
end;

function TXmlEditPageControl.ActivatePage(const SysId: WideString): Boolean;
// Activates the page if the file is already open.
var
  I: Integer;
begin
  I := IndexOfPage(SysId);
  if I > -1 then begin
    ActivePage := Pages[I];
    Result := True;
  end else
    Result := False;
end;

procedure TXmlEditPageControl.AddEmptyDoc;
var
  NewTabSheet: TXmlEditTabSheet;
begin
  // Create and add a new editor tab sheet:
  NewTabSheet := TXmlEditTabSheet.Create(Self, nil, '', False, -1, EditorFont, ExternalHexEditor);
  NewTabSheet.PageControl := Self;
  ActivePage := NewTabSheet;
end;

function TXmlEditPageControl.CanCheckValidity: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanCheckValidity
    else Result := False;
end;

function TXmlEditPageControl.CanCheckWellformedness: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanCheckWellformedness
    else Result := False;
end;

function TXmlEditPageControl.CanCloseAll: Boolean;
var
  I: Integer;
begin
  I := Pred(PageCount);
  while I >= 0 do begin
    case Pages[I].AskSaveChanges of
      IDYes:
        Result := SavePage(Pages[I]);
      IDNo:
        Result := True;
    else
      Result := False;
    end;
    if not Result then
      Exit;
    Dec(I);
  end;
  Result := True;
end;

function TXmlEditPageControl.CanCopy: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanCopy
    else Result := False;
end;

function TXmlEditPageControl.CanCut: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanCut
    else Result := False;
end;

function TXmlEditPageControl.CanDelete: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanDelete
    else Result := False;
end;

function TXmlEditPageControl.CanPaste: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanPaste
    else Result := False;
end;

function TXmlEditPageControl.CanPrint: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanPrint
    else Result := False;
end;

function TXmlEditPageControl.CanRedo: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanRedo
    else Result := False;
end;

function TXmlEditPageControl.CanReplace: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanReplace
    else Result := False;
end;

function TXmlEditPageControl.CanSave: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanSave
    else Result := False;
end;

function TXmlEditPageControl.CanSaveAs: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanSaveAs
    else Result := False;
end;

function TXmlEditPageControl.CanSearch: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanSearch
    else Result := False;
end;

function TXmlEditPageControl.CanSelectAll: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanSelectAll
    else Result := False;
end;

function TXmlEditPageControl.CanUndo: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CanUndo
    else Result := False;
end;

function TXmlEditPageControl.CaretPosStr: string;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CaretPosStr
    else Result := '';
end;

function TXmlEditPageControl.CheckValidity: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CheckValidity
    else Result := True;
end;

function TXmlEditPageControl.CheckWellformedness: Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.CheckWellformedness
    else Result := True;
end;

procedure TXmlEditPageControl.Close;

  procedure PostClosingMessage;
  // Remark: Instead of freeing the active page directly, we put a
  //         message at the end of the windows message queue.  This
  //         message is executed after all other pending messages.
  //         This makes sure that an event handler of the active page
  //         which called TXmlEditPageControl.Close has finished before
  //         the active pages is freed.
  begin
    PostMessage(Self.Handle, WM_CLOSE_TAB, Integer(ActivePage), 0);
  end;

begin
  if Assigned(ActivePage) then begin
    case ActivePage.AskSaveChanges of
      IDYes:
        if SavePage(ActivePage) then
          PostClosingMessage;
      IDNo:
        PostClosingMessage;
    end;
    // Remark: If the cancel button was pressed or saving failed, no closing
    //         message is posted.
  end;
end;

procedure TXmlEditPageControl.CloseAll;
var
  I: Integer;
begin
  I := Pred(PageCount);
  while I >= 0 do begin
    PostMessage(Self.Handle, WM_CLOSE_TAB, Integer(Pages[I]), 0);
      // Remark: Instead of freeing the active page directly, we put a
      //         message at the end of the windows message queue.  This
      //         message is executed after all other pending messages.
      //         This makes sure that an event handler of the active page
      //         which called TXmlEditPageControl.Close has finished before
      //         the active pages is freed.
    Dec(I);
  end;
end;

constructor TXmlEditPageControl.Create(AOwner: TComponent);
var
  AFont: TFont;
begin
  inherited;
  FSaxonProcessCount := 0;

  FEditorFont := TFont.Create;

  FErrorStringIterator := TUtilsStringIterator.Create(Self);

  FSaveDialog := TSaveDialog.Create(Self);
  FSaveDialog.Options := [ofOverwritePrompt, ofHideReadOnly, ofEnableSizing];

  FPageSetupDlg := TPageSetupDlg.Create(Self);
  FPrintDialog := TPrintDialog.Create(Self);
  FSynEditPrint := TSynEditPrint.Create(Self);
  FXmlHighlighter := TSynXmlSyn.Create(Self);
  FXmlHighlighter.Name := 'XmlHighlighter';
  FXmlHighlighter.SetSubComponent(True);

  // Initialize FSynEditPrint:
  AFont := TFont.Create;
  try
    with FSynEditPrint.Header do begin
        {First line, default font, left aligned}
      Add('$TITLE$', nil, taLeftJustify, 1);
        {First line, default font, right aligned}
      Add('Page: $PAGENUM$ of $PAGECOUNT$', nil, taRightJustify, 1);
    end;
    with FSynEditPrint.Footer do begin
      AFont.Assign(DefaultFont);
      AFont.Size := 6;
      {First line, small font, left aligned - note that lines can have different fonts}
      Add('Printed by Open XML Editor', AFont, taLeftJustify, 1);
      {First line, small font, right aligned}
      Add('Print Date: $DATE$. Time: $TIME$', AFont, taRightJustify, 1);
    end;
  finally
    AFont.Free;
  end;

  // Initialize the Close Button:
  FCloseBtn := TSpeedButton.Create(Self);
  with FCloseBtn do begin
    Anchors := [akTop, akRight];
    Flat := True;
    Glyph.LoadFromResourceName(HInstance, 'CLOSEBTNGLYPH');
    Height := 22;
    Parent := nil;
    ParentShowHint := False;
    ShowHint := True;
    Top := 2;
    Width := 23;
    OnClick := CloseBtnClickHandler;
  end;

  // Initialize Options:
  FStartEmpty := True;
end;

destructor TXmlEditPageControl.Destroy;
begin
  FEditorFont.Free;
  inherited;
end;

procedure TXmlEditPageControl.CloseBtnClickHandler(Sender: TObject);
begin
  Close;
end;

procedure TXmlEditPageControl.DoSaved(const OldUri, NewUri: WideString);
begin
  if Assigned(FOnSaved) then
    FOnSaved(Self, OldUri, NewUri);
end;

procedure TXmlEditPageControl.DoSaxonLaunching;
begin
  if Assigned(FOnSaxonLaunching) then
    FOnSaxonLaunching(Self);
end;

procedure TXmlEditPageControl.DoSearchReplaceText(AReplace,
  ABackwards: Boolean);
// Modified version of code taken from the EditAppDemos
// project, written by Michael Hieke for the SynEdit
// component suite (see "http://SynEdit.SourceForge.net").
var
  Options: TSynSearchOptions;
begin
  if not Assigned(ActivePage) then Exit;
  if AReplace then
    Options := [ssoPrompt, ssoReplace, ssoReplaceAll]
  else
    Options := [];
  if ABackwards then
    Include(Options, ssoBackwards);
  if FSearchParams.SearchCaseSensitive then
    Include(Options, ssoMatchCase);
  if not FSearchParams.SearchFromCaret then
    Include(Options, ssoEntireScope);
  if FSearchParams.SearchSelectionOnly then
    Include(Options, ssoSelectedOnly);
  if FSearchParams.SearchWholeWords then
    Include(Options, ssoWholeWord);
  ActivePage.SearchReplace(FSearchParams.SearchText, FSearchParams.ReplaceText, Options);
end;

procedure TXmlEditPageControl.DoSetupHexEditorRequest;
begin
  if Assigned(FOnSetupHexEditorRequest) then
    FOnSetupHexEditorRequest(Self);
end;

procedure TXmlEditPageControl.DoTabChange;
begin
  if Assigned(ActivePage) then begin
    ErrorStringIterator.List := ActivePage.ErrorStrings;
    FCloseBtn.Parent := Self;
    FCloseBtn.Left := FCloseBtn.Parent.Width - FCloseBtn.Width - 2;
    CheckRefreshActivePage;
  end else begin
    ErrorStringIterator.List := nil;
    FCloseBtn.Parent := nil;
  end;
  if Assigned(FOnTabChange) then FOnTabChange(Self);
end;

procedure TXmlEditPageControl.FocusEditor;
begin
  if Assigned(ActivePage) then
    ActivePage.FocusEditor;
end;

function TXmlEditPageControl.GetActivePage: TXmlEditTabSheet;
begin
  Result := inherited ActivePage as TXmlEditTabSheet;
end;

function TXmlEditPageControl.GetActivePageIndex: Integer;
begin
  Result := inherited ActivePageIndex;
end;

function TXmlEditPageControl.GetEditorFont: TFont;
begin
  Result := FEditorFont;
end;

function TXmlEditPageControl.GetExternalHexEditor: TFilename;
begin
  Result := FExternalHexEditor;
end;

function TXmlEditPageControl.GetPage(Index: Integer): TXmlEditTabSheet;
begin
  Result := inherited Pages[Index] as TXmlEditTabSheet;
end;

function TXmlEditPageControl.GetSaxon: TFilename;
begin
  Result := FSaxon;
end;

function TXmlEditPageControl.GetUntitledNumber: Integer;
// Code taken from the EditAppDemos project, written by
// Michael Hieke for the SynEdit component suite
// (see "http://SynEdit.SourceForge.net").
begin
  if FUntitledNumbers = nil then
    FUntitledNumbers := TBits.Create;
  Result := FUntitledNumbers.OpenBit;
  if Result = FUntitledNumbers.Size then
    FUntitledNumbers.Size := FUntitledNumbers.Size + 32;
  FUntitledNumbers[Result] := TRUE;
  Inc(Result);
end;

function TXmlEditPageControl.HasActivePage: Boolean; 
begin
  Result := Assigned(ActivePage);
end;

function TXmlEditPageControl.SaxonIsBusy: Boolean;
begin
  Result := FSaxonProcessCount > 0;
end;

procedure TXmlEditPageControl.AddDoc(const Data: TBytes;
                                     const SysId: WideString;
                                     const Modified: Boolean);
// If the specified system identifier is empty, then add a new page containing
// the specified data.
// Else: if a page with the specified system identifier already exists, then
// replace the text of the page with S.
// Else: if the active page is empty, then replace the text of the page with S.
// Else: add a new page containing the specified data.
// Parameters:
//   Data:     The new text of the page.
//   SysId:    The system identifier of the file where S comes from.
//   Modified: If true, marks the page as modified.

  function GetTimeStamp(const SysId: WideString): TDateTime;
  // If the resource specified in SysId is local, obtain OS timestamp,
  // otherwise return -1.
  var
    Path: TFileName;
    Authority, Query, Fragment: WideString; // dummies
  begin
    UriWideStrToFilename(SysId, Path, Authority, Query, Fragment);
    if not FileAge(Path, Result) then
      Result := -1;
  end;

  function AddNewTabSheet(const Data: TBytes;
                          const SysId: WideString;
                          const Modified: Boolean;
                          const FileAge: TDateTime): TXmlEditTabSheet;
  begin
    Result := TXmlEditTabSheet.Create(Self, Data, SysId, Modified, FileAge, EditorFont, ExternalHexEditor);
    Result.PageControl := Self;
    ActivePage := Result;
  end;

var
  I: Integer;
  TimeStamp: TDateTime;
begin
  // Obtain OS timestamp (or -1 if the resource is not local, i.e. if it is not of type "file://"):
  TimeStamp := GetTimeStamp(SysId);

  I := IndexOfPage(SysId);
  if (Trim(SysId) = '') or (I = -1) then begin

    // Add new page:
    if Assigned(ActivePage) then begin
      if ActivePage.IsEmpty and (ActivePage.DocumentUri = '') then begin
        ActivePage.DocumentUri := SysId;
        ActivePage.TimeStamp := TimeStamp;
        ActivePage.LoadFromBytes(Data, Modified);  // Remark: LoadFromBytes() must be go after the TimeStamp assignment, because
                                                   //         the time stamp is checked when switch to HexEditor is required
                                                   //         in LoadFromBytes().
                                                   { TODO 5 -cRefactor : Check Remark! }
      end else AddNewTabSheet(Data, SysId, Modified, TimeStamp);
    end else AddNewTabSheet(Data, SysId, Modified, TimeStamp);

  end else begin

    // Reload page:
    Pages[I].LoadFromBytes(Data, Modified);
    Pages[I].TimeStamp := TimeStamp;
    ActivePage := Pages[I];

  end;
end;

// I/O operations

procedure TXmlEditPageControl.CheckRefreshActivePage;
const
  MBType = MB_YESNO or MB_ICONQUESTION;
var
  TimeStamp: TDateTime;
  S: string;
begin
  if Assigned(ActivePage) then begin
    if FileAge(ActivePage.FileName, TimeStamp) then
      if TimeStamp <> ActivePage.TimeStamp then begin
        S := Format(SAskReload, [ActivePage.DocumentUri]);
        if Application.MessageBox(PChar(S), PChar(Application.Title), MBType) = IDYes
          then Reload;
        ActivePage.TimeStamp := TimeStamp;
      end;
  end;
end;

function TXmlEditPageControl.RequestSaveFileName(var AFilename: TFilename): Boolean;
begin
  with FSaveDialog do begin
    if AFilename = '' then begin
      InitialDir := '';
      Filename := '';
    end else begin
      InitialDir := ExtractFileDir(AFilename);
      FileName := ExtractFileName(AFilename);
    end;
    Filter := SFilterMaskXmlFiles + '|' + SFilterMaskAllFiles;
    if Execute then begin
      AFilename := FileName;
      Result := True;
    end else Result := False;
  end;
end;

function TXmlEditPageControl.OpenUri(PubId,
                                     SysId: WideString): Boolean;

var
  Data: TBytes;
begin
  try
    Data := ResolveResource(PubId, SysId);  // Retrieve resource as string.
    AddDoc(Data, SysId, False);             // Add new document tab sheet.
    Result := True;
  except
    Result := False;
  end;
end;

function TXmlEditPageControl.Reload: Boolean;
const
  MBType = MB_YESNO or MB_ICONQUESTION;
var
  Data: TBytes;
  PubId, SysId: WideString;
  TimeStamp: TDateTime;
begin
  if Assigned(ActivePage) then begin
    Result := True;
    PubId := '';
    SysId := ActivePage.DocumentUri;
    try
      Data := ResolveResource(PubId, SysId);
      ActivePage.LoadFromBytes(Data, False);
      if not FileAge(ActivePage.FileName, TimeStamp) then
        TimeStamp := -1;
      ActivePage.TimeStamp := TimeStamp;
    except
      Result := False;
    end;
  end else
    Result := False;
end;

procedure TXmlEditPageControl.Save;
var
  OldUri: WideString;
  TimeStamp: TDateTime;
begin
  if Assigned(ActivePage) then begin
    OldUri := ActivePage.DocumentUri;
    if SavePage(ActivePage) then begin
      if not FileAge(ActivePage.FileName, TimeStamp) then
        TimeStamp := -1;
      ActivePage.TimeStamp := TimeStamp;
      DoSaved(OldUri, ActivePage.DocumentUri);
    end;
  end;
end;

procedure TXmlEditPageControl.SaveAs;
var
  OldUri: WideString;
  TimeStamp: TDateTime;
begin
  if Assigned(ActivePage) then begin
    OldUri := ActivePage.DocumentUri;
    if SavePageAs(ActivePage) then begin
      if not FileAge(ActivePage.FileName, TimeStamp) then
        TimeStamp := -1;
      ActivePage.TimeStamp := TimeStamp;
      DoSaved(OldUri, ActivePage.DocumentUri);
    end;
  end;
end;

function TXmlEditPageControl.SavePage(const Page: TXmlEditTabSheet): Boolean;
begin
  if Page.FileName = ''
    then Result := SavePageAs(Page)
    else Result := Page.Save;
end;

function TXmlEditPageControl.SavePageAs(const Page: TXmlEditTabSheet): Boolean;

  function IsOpenInAnotherPage(const Page: TXmlEditTabSheet;
                               const SysId: WideString): Boolean;
  var
    I: Integer;
  begin
    I := IndexOfPage(SysId);
    if I > -1
      then Result := Pages[I] <> Page
      else Result := False;
  end;

const
  MBType = MB_OK or MB_ICONEXCLAMATION;
var
  Path: TFilename;
  SysId: WideString;
  SysIdUsedElsewhere: Boolean;
begin
  Path := Page.FileName;

  SysIdUsedElsewhere := False;
  repeat
    if SysIdUsedElsewhere then
      Application.MessageBox(PChar(SNameAlreadyInuse), PChar(Application.Title), MBType);
    if not RequestSaveFileName(Path) then begin
      Result := False;
      Exit;
    end;
    SysId := FilenameToUriWideStr(Path, []);
    SysIdUsedElsewhere := IsOpenInAnotherPage(Page, SysId);
  until not SysIdUsedElsewhere;

  Page.DocumentUri := SysId;
  Result := Page.Save;
end;

procedure TXmlEditPageControl.Notification(AComponent: TComponent; operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) then begin
    if AComponent = FResourceResolver then
      FResourceResolver := nil
    else if AComponent = FPrinterSetupAction then
      FPrinterSetupAction := nil;
  end;
end;

procedure TXmlEditPageControl.ReleaseUntitledNumber(ANumber: Integer);
// Code taken from the EditAppDemos project, written by
// Michael Hieke for the SynEdit component suite
// (see "http://SynEdit.SourceForge.net").
begin
  Dec(ANumber);
  if (FUntitledNumbers <> nil) and (ANumber >= 0)
    and (ANumber < FUntitledNumbers.Size)
  then
    FUntitledNumbers[ANumber] := FALSE;
end;

function TXmlEditPageControl.ResolveResource(var PublicId,
                                                 SystemId: WideString): TBytes;

  function StreamToBytes(Stream: TStream): TBytes;
  var
    Size: Integer;
  begin
    Size := Stream.Size - Stream.Position;
    SetLength(Result, Size);
    Stream.Read(Pointer(Result)^, Size);
  end;

var
  InputStream: TStream;
begin
  if not Assigned(ResourceResolver) then
    raise EFOpenError.CreateFmt(SFOpenError, []);

  InputStream := ResourceResolver.ResolveResource('', PublicId, SystemId);
  try
    Result := StreamToBytes(InputStream);
  finally
    InputStream.Free;
  end;
end;

procedure TXmlEditPageControl.SetActivePage(const Value: TXmlEditTabSheet);
begin
  if Value <> GetActivePage then begin
    inherited SetActivePage(Value);
    DoTabChange;
  end;
end;

procedure TXmlEditPageControl.SetActivePageIndex(const Value: Integer);
begin
  if Value <> GetActivePageIndex then begin
    inherited ActivePageIndex := Value;
    DoTabChange;
  end;
end;

function TXmlEditPageControl.SetCaret(const Location: IDomLocator): Boolean;
begin
  if Assigned(ActivePage)
    then Result := ActivePage.SetCaret(Location)
    else Result := False;
end;

procedure TXmlEditPageControl.SetEditorFont(const Value: TFont);
var
  I: Integer;
begin
  if FEditorFont <> Value then begin
    FEditorFont.Assign(Value);
    for I := 0 to Pred(PageCount) do
      Pages[I].EditorFont := Value;
  end;
end;

procedure TXmlEditPageControl.SetExternalHexEditor(const Value: TFilename);
var
  I: Integer;
begin
  if FExternalHexEditor <> Value then begin
    FExternalHexEditor := Value;
    for I := 0 to Pred(PageCount) do
      Pages[I].ExternalHexEditor := Value;
  end;
end;

procedure TXmlEditPageControl.SetSaxon(const Value: TFilename);
begin
  FSaxon := Value;
end;

procedure TXmlEditPageControl.ShowControl(AControl: TControl);
begin
  if (AControl is TXmlEditTabSheet) then
    if (TXmlEditTabSheet(AControl).PageControl = Self) and (ActivePage <> TTabSheet(AControl)) then
      DoTabChange;
  inherited;
end;

procedure TXmlEditPageControl.SetStartEmpty(const Value: Boolean);
begin
  FStartEmpty := Value;
end;

procedure TXmlEditPageControl.UpdateActivePage;
var
  OldActivePage: TXmlEditTabSheet;
begin
  FInSetActivePage := True;
  OldActivePage := ActivePage;
  inherited;
  if ActivePage <> OldActivePage then
    DoTabChange;
  FInSetActivePage := False;
end;

procedure TXmlEditPageControl.WndProc(var Message: TMessage);
var
  EditTabSheet: TXmlEditTabSheet;
begin
  if Message.Msg = CM_ENTER then begin // WM_SETFOCUS then begin

    CheckRefreshActivePage;
    inherited WndProc(Message);

  end else if Message.Msg = WM_CLOSE_TAB then begin

    EditTabSheet := TXmlEditTabSheet(Message.WParam);
    if FCloseBtn.Parent = Self then     { TODO 5 -cRefactor : Revisit! }
      FCloseBtn.Parent := nil;
    EditTabSheet.Free;

    if (PageCount = 0) and StartEmpty then
      AddEmptyDoc;

  end else

    inherited WndProc(Message);
end;

// Tools operations

procedure TXmlEditPageControl.LaunchSaxon;
var
  Path: TFilename;
  Authority, Query, Fragment: WideString; // dummies
  SaxonLauncher: TAppLauncher;
  SaxonExitCode: Cardinal;
  StdOut: TBytes;
  StdErr: TBytes;
  XsltOutputDialog: TXsltOutputDlg;
  XsltErrorDlg: TXsltErrorDlg;
  UTF8Codec: TUTF8Encoding;
  Stream: TStream;
begin
  if Assigned(ActivePage) then begin

    if ActivePage.IsEmpty then begin
      MessageBeep(MB_ICONEXCLAMATION);
      ShowMessage(SDocumentIsEmpty);
    end else if ActivePage.Modified then begin
      MessageBeep(MB_ICONEXCLAMATION);
      ShowMessage(SMustBeSaved);
    end else begin
      SaxonLauncher := TAppLauncher.Create(nil);
      try
        SaxonLauncher.CreateNoWindow := True;

        Inc(FSaxonProcessCount);
        try

          // Launch Saxon.
          DoSaxonLaunching;
          UriWideStrToFilename(ActivePage.DocumentUri, Path, Authority, Query, Fragment);
          SaxonExitCode := SaxonLauncher.ExecuteW(Saxon,
            Format('"%s" -s:"%s" -a -warnings:silent"',
                   [Saxon, Path]), StdOut, StdErr);

        finally
          Dec(FSaxonProcessCount);
        end;

        if (SaxonExitCode = 0) then begin
          // XSLT processing succeeded.

          XsltOutputDialog := TXsltOutputDlg.Create(Self);
          try
            case XsltOutputDialog.ShowModal of
              mrOK:
                // Open in editor.
                AddDoc(StdOut, '', True);
              mrYes: begin
                // Save to file
                Stream := TFileStream.Create(XsltOutputDialog.SaveDialog.FileName, fmCreate);
                try
                  Stream.WriteBuffer(StdOut[0], Length(StdOut));
                finally
                  Stream.Free;
                end;
              end;
            end;
          finally
            XsltOutputDialog.Free;
          end;

        end else begin
          // XSLT processing failed.

          XsltErrorDlg := TXsltErrorDlg.Create(nil);
          try
            UTF8Codec := TUTF8Encoding.Create;
            try
              XsltErrorDlg.OutputMemo.Text := UTF8Codec.GetString(StdErr);
            finally
              UTF8Codec.Free;
            end;
            MessageBeep(MB_ICONEXCLAMATION);
            XsltErrorDlg.ShowModal;
          finally
            XsltErrorDlg.Free;
          end;
        end;

      finally
        SaxonLauncher.Free;
      end;
    end;

  end;
end;

initialization

  WM_CLOSE_TAB := RegisterWindowMessage('WM_CLOSE_TAB');
    // Remark: RegisterWindowMessage makes sure that the message IDs are unique
    //         on the system.

end.
